#!/usr/bin/python
# $Id: download.py,v 1.19 2004/03/09 08:44:02 torh Exp $

#  Copyright 2001 - 2003 Trustix AS - <http://www.trustix.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import sys, time, os, urlparse, string, re, shutil
from swuplib import gpg
from swuplib import log
import urllib

DEBUG = 0

# some constants/defaults
CACHEDIR = '/var/cache/swup'
TMPDIR = '/var/spool/swup/tmp'
GNUPGDIR='/etc/swup/gnupg'
SIGNATURE_APPEND = '.asc'
false, true = 0, 1

# define some exceptions
class DownloadException(Exception): pass
class FlushCacheException(DownloadException): pass
class SignatureException(DownloadException): pass
class AuthException(DownloadException): pass

class Flush:
    def __init__( self, configdict):
        "Set up configuration and run initialization functions."
        self.config = configdict
	self.cachedir = configdict['cachedir']
        self.log = log.log(self.config["loglevel"], False )
    
    def run ( self, site=None ):
	urllib.urlcleanup()
	try:
	    if os.path.isdir(self.cachedir):
                for subdir in ['rpms','rdfs','lists']:
                    dir = os.path.join(self.cachedir, subdir)
                    if site:
                        dir = os.path.join(dir, site)
                    if os.path.isdir(dir):
                        shutil.rmtree(dir)
                    if not site:
                        os.mkdir(dir, 0755)
                if not site:
                    #
                    # Now let'a recreate the infrastructure:
                    #
                    for subdir in ['file:', 'http:', 'https:', 'ftp:']:
                        for type in ['rpms/','rdfs/']:
                            sd = type + subdir
                            dir = os.path.join(self.cachedir, sd)
                            os.mkdir(dir, 0755)
            else:
                errmsg = "Unable to flush cache: '%s' Not a directory." \
                            %self.cachedir
                sys.stderr.write("%s\n" %errmsg)
                self.log.write_syslog_err(errmsg)
                
        except:
            raise



def byte_convert( bytes ):
    """Returns a two tuple (<str>, <prefix>), where prefix is one of
    '', 'k' or 'M'."""
    div = bytes/float(1024)
    if div < 1:
        return (str(bytes), '')
    if div < 1024:
        return ( "%4.2f" %div, 'k')
    else:
        div = bytes / float(1024**2)
        return ("%4.2f" %div, 'M')



class download:

    def __init__( self, gnupgdir=GNUPGDIR, log=None, cachedir=CACHEDIR, tmpdir=TMPDIR):
        self.log = log
	self.cachedir = cachedir
	self.tmpdir = tmpdir
        self.gnupgdir = gnupgdir
        self.auth_handler = None
	try:
            import urllib2
	    self.opener = urllib2
            self.auth_handler = urllib2.HTTPBasicAuthHandler( \
	        urllib2.HTTPPasswordMgrWithDefaultRealm())
	    handler = self.opener.build_opener(self.auth_handler,\
			    urllib2.CacheFTPHandler())
	    self.opener.install_opener(handler)
	except ImportError:
	    self.opener = urllib
	except:
	    raise
	
	if not os.path.isdir( tmpdir ):
            try:
                if os.path.exists( tmpdir ):
                    os.remove( tmpdir )
                os.makedirs( tmpdir )
            except (OSError, IOError), err:
                raise DownloadException, err
            except:
                raise


    def callback(self, blocknum, blocksize, filesize):
        #if size == -1: don't use size in calculation.

        if blocknum == 0:
            self._bytes_read = 0
            self._time_used = 0
            self._time = time.time()

        self._bytes_read = self._bytes_read + blocksize

        if filesize not in [-1,0]:
            if self._bytes_read > filesize:
                percent = 100
                self._bytes_read = filesize
            else:
                tmptime = time.time()
                self._time_used = self._time_used + (tmptime - self._time)
                self._time = tmptime
                percent = self._bytes_read / float(filesize) * 100
            if not self._time_used <= 0:
                (speed, speed_prefix) = byte_convert( self._bytes_read/
                                                      float(self._time_used))
            else:
                speed, speed_prefix = 0, 'k'
            
            (size, size_prefix) = byte_convert( filesize )
            (read, read_prefix) = byte_convert( self._bytes_read )
            message = "\rDownloading "\
                      "%-24s (%-7s%-1sb of %-7s%-1sb, %-3d%%, %-7s%-1sb/s)"%\
                      (self._now_downloading[:21]+"...", read, read_prefix,
                       size, size_prefix, percent, speed, speed_prefix)
        else:
            (read, read_prefix) = byte_convert( self._bytes_read )
            message = "\rDownloading %-56s (%-7s%-1sb)" %\
                      (self._now_downloading[:27], read, read_prefix)
        
        if self.log != None:
            self.log.write_tty(message)


    def _do_download( self, uri ):
        scheme, host, path, param, query, fragment = urlparse.urlparse( uri )
        if '@' in host and self.auth_handler and scheme in ['http', 'https']:
            try:
                user_password, host = string.split(host, '@', 1)
                user, password = string.split(user_password, ':', 1)
                self.auth_handler.add_password(None, host, user, password)
            except ValueError, e:
	        raise AuthException, ('Bad URL: %s') % url 
        elif '@' in host and scheme == 'ftp':
            raise AuthException, 'FTP Authentication not supported.'

        uri = urlparse.urlunparse((scheme, host, path, param, query, fragment))

        # Make sure the cache is ready to accept the file.
	filename = os.path.normpath(os.path.join(self.cachedir, uri))
        
        basename = os.path.basename( uri )
        self._now_downloading = basename
	tmpfilename = os.path.join( self.tmpdir, basename )

        try:
            urlobj = self.opener.urlopen(uri)
            
            # Write the data to file named 'tmpfilename':
            fileobj = open(tmpfilename, 'wb')
            blocksize = 1024*8
            block = urlobj.read(blocksize)
            while block:
                fileobj.write(block)
                block = urlobj.read(blocksize)
            fileobj.close()
            # WTF ?? Return before checking if we should compress ? 
            #return tmpfilename

            # gunzip compressed files
            if tmpfilename[-3:] == '.gz':
                import gzip
                gfd = gzip.open(tmpfilename)
                localfilename = tmpfilename[:-3]
                fd = open(localfilename, 'wb')
                block = gfd.read(8192)
                while block:
                    fd.write(block)
                    block = gfd.read(8192)
                gfd.close()
                fd.close()
                del(gfd, fd)
                os.remove(tmpfilename)
            else:
                localfilename = tmpfilename
            return localfilename
        except IOError, err:
	    raise DownloadException, err
        except:
            raise


    def download( self, url, sigurl=None, CHECKSIG=1, TYPE=1):
	"""Download a file and signature, check signature, move to cachedir.

	The argument is the url of the file. Optional argument is the url
	of the signature. The default value for the signature name is
	url.asc
        
        TYPE tells us where to put the cached files. 
        1 = rpms
        2 = rdfs
        """

        url = string.strip(url)
	# Check if the file exists - if it does, return
        if TYPE == 1:
            localurl = "rpms/" + url
        elif TYPE == 2:
            localurl = "rdfs/" + url
	if urlparse.urlparse(url)[0] == "":
	    url = "file:" + url
            
	filename = os.path.normpath(os.path.join(self.cachedir, localurl))
	cachepath = os.path.dirname(filename)

	if os.path.exists(filename):
	    # if we already have the file, don't download again...
	    return filename

	if not os.path.isdir(cachepath):
	    if os.path.exists(cachepath):
		os.remove(cachepath)
            os.makedirs(cachepath)

	# Actually download stuff
	if sigurl == None:
            if url[-3:] == '.gz':
                sigurl = url[:-3] + SIGNATURE_APPEND
            else:
                sigurl = url + SIGNATURE_APPEND

        # Get the signature
        if CHECKSIG:
            try:
                signame = self._do_download( sigurl )
            except AuthException, errmsg:
                host = urlparse.urlparse(sigurl)[1]
                raise AuthException, "\nAuthentication failed for host %s:\n %s\n" % (host, errmsg)
            except DownloadException, errmsg:
               raise DownloadException, "\nDownload failed for %s:\n %s\n" % (sigurl, errmsg)
            except:
                raise

        # Get the file
        try:
            tmpfilename = self._do_download( url )
        except AuthException, errmsg:
            host = urlparse.urlparse(url)[1]
            raise AuthException, "\nAuthentication failed for host %s:\n %s\n" % (host, errmsg)
        except DownloadException, errmsg:
            raise DownloadException, "\nDownload failed for %s:\n %s\n" % (url, errmsg)
        except:
            raise

	# Check the sig
        if CHECKSIG and tmpfilename and signame:
            try:
                retval, output = gpg.checksig(signame, self.gnupgdir)
                os.remove(signame)
            except gpg.signature_fail:
		errmsg = \
"The PGP signature of this file could not be verified, possible issues are:\n"+\
"- The PGP key which this file is signed by, is not yet imported into swup.\n"+\
"- The systems clock is set horribly wrong.\n"+\
"- The file, or its signature file, was damaged during download.\n"+\
"- The file was not signed at all.\n"+\
"In any case, this file will not be trusted.\n"
                raise SignatureException, errmsg
            except:
                raise
            
	# If we have come this far, we should have the correct file properly
	# downloaded. Time to move it...
        if tmpfilename:
            shutil.copy(tmpfilename, filename)
            os.remove(tmpfilename)
        else:
            filename = None


	# return just the filename - we don't really care about the output...
	return filename

